
#include <windows.h>
#include <string.h>

#define OTHEXPORT __declspec(dllexport)

#include "oh_trmod.h"
#include "ohbbhlp.h"

/* keep BUFFSIZE small because of translation software bug. */
#define BUFFSIZE 1024
#define OTHPATHLEN 2048

#define OTHLOCKINIT() InitializeCriticalSection(&CriticalSection)
#define OTHLOCK() EnterCriticalSection(&CriticalSection)
#define OTHUNLOCK() LeaveCriticalSection(&CriticalSection)
#define OTHLOCKDEL() DeleteCriticalSection(&CriticalSection)

static const OHBSETTING *pTransSetting;
static HINSTANCE hInst;
static HMODULE hBabel;
static int isRunning;

static CRITICAL_SECTION CriticalSection;
static WCHAR jpbuff[BUFFSIZE];
static int jpendpos;
static WCHAR krbuff[BUFFSIZE];
static int krendpos;
static HANDLE jktrth;

static void OthStrStr(LPWSTR s2, LPWSTR *s3);
static LPWSTR OthJKTrans(LPWSTR jpStr);
static void OthPushJP(LPCWSTR jpStr, int len);
static LPWSTR OthPullJP(void);
static void OthPushKR(LPWSTR krStr);
static LPWSTR OthPullKR(void);

static void OthJKTransThread(void *arg);

static int (OHBCALL *pfnOhbInit)(OHBTRANSID iTransID);
static void (OHBCALL *pfnOhbDestroy)(OHBTRANSID iTransID);
static const OHBSETTING * (OHBCALL *pfnOhbGetSettings)(int *piSetting);
static const OHBTRANS * (OHBCALL *pfnOhbTrans)(LPCWSTR jpStr,
  const OHBSETTING *pSetting);
static void (OHBCALL *pfnOhbFree)(const OHBTRANS *ohbTrans);

/* hack to optimize */
static void OthStrStr(LPWSTR s2, LPWSTR *s3)
{
  LPWSTR s1;
  for (s1 = jpbuff + jpendpos; s1 != jpbuff && s1 > *s3; --s1)
  {
    if (*s1 == *s2)
    {
      *s3 = s1;
    }
  }
}

static LPWSTR OthJKTrans(LPWSTR jpStr)
{
  int len;
  LPWSTR krStr;
  const OHBTRANS *tr;

  tr = pfnOhbTrans(jpStr, pTransSetting);

  len = lstrlenW(tr->lpStr);
  krStr = malloc((len + 1) * 2);
  lstrcpyW(krStr, tr->lpStr);

  pfnOhbFree(tr);

  return krStr;
}

static void OthPushJP(LPCWSTR jpStr, int len)
{
  int i;

  if (len <= 0)
    return;

  OTHLOCK();

  if (jpendpos >= BUFFSIZE)
  {
    jpbuff[0] = 0;
    jpendpos = 0;
  }

  for (i = 0; i < len && jpendpos < BUFFSIZE && jpStr[i] != 0;
    ++i, ++jpendpos)
  {
    jpbuff[jpendpos] = jpStr[i];
  }

  if (jpendpos >= BUFFSIZE)
  {
    jpbuff[0] = 0;
    jpendpos = 0;
  }

  jpbuff[jpendpos] = 0;

  OTHUNLOCK();
}

static LPWSTR delimiters[] =
  {
    (LPWSTR)"\x21\x00",
    (LPWSTR)"\x22\x00",
    (LPWSTR)"\x27\x00",
    (LPWSTR)"\x28\x00",
    (LPWSTR)"\x29\x00",
    (LPWSTR)"\x2c\x00",
    (LPWSTR)"\x2e\x00",
    (LPWSTR)"\x3a\x00",
    (LPWSTR)"\x3b\x00",
    (LPWSTR)"\x3c\x00",
    (LPWSTR)"\x3e\x00",
    (LPWSTR)"\x3f\x00",
    (LPWSTR)"\x5b\x00",
    (LPWSTR)"\x5d\x00",
    (LPWSTR)"\x5e\x00",
    (LPWSTR)"\x60\x00",
    (LPWSTR)"\x7b\x00",
    (LPWSTR)"\x7d\x00",
    (LPWSTR)"\xab\x00",
    (LPWSTR)"\xbb\x00",
    (LPWSTR)"\xb4\x00",
    (LPWSTR)"\xb7\x00",
    (LPWSTR)"\xb8\x00",
    (LPWSTR)"\x01\x30",
    (LPWSTR)"\x02\x30",
    (LPWSTR)"\x0c\xff",
    (LPWSTR)"\x0e\xff",
    (LPWSTR)"\xfb\x30",
    (LPWSTR)"\x1a\xff",
    (LPWSTR)"\x1b\xff",
    (LPWSTR)"\x1f\xff",
    (LPWSTR)"\x01\xff",
    (LPWSTR)"\x9b\x30",
    (LPWSTR)"\x9c\x30",
    (LPWSTR)"\xb4\x00",
    (LPWSTR)"\x40\xff",
    (LPWSTR)"\xa8\x00",
    (LPWSTR)"\x3e\xff",
    (LPWSTR)"\x18\x20",
    (LPWSTR)"\x19\x20",
    (LPWSTR)"\x1c\x20",
    (LPWSTR)"\x1d\x20",
    (LPWSTR)"\x08\xff",
    (LPWSTR)"\x09\xff",
    (LPWSTR)"\x14\x30",
    (LPWSTR)"\x15\x30",
    (LPWSTR)"\x3b\xff",
    (LPWSTR)"\x3d\xff",
    (LPWSTR)"\x5b\xff",
    (LPWSTR)"\x5d\xff",
    (LPWSTR)"\x08\x30",
    (LPWSTR)"\x09\x30",
    (LPWSTR)"\x0a\x30",
    (LPWSTR)"\x0b\x30",
    (LPWSTR)"\x0c\x30",
    (LPWSTR)"\x0d\x30",
    (LPWSTR)"\x0e\x30",
    (LPWSTR)"\x0f\x30",
    (LPWSTR)"\x10\x30",
    (LPWSTR)"\x11\x30",
    (LPWSTR)"\x1c\xff",
    (LPWSTR)"\x1e\xff",
    (LPWSTR)"\x32\x20",
    (LPWSTR)"\x33\x20",
    (LPWSTR)"\x6a\x22",
    (LPWSTR)"\x6b\x22",
    (LPWSTR)"\x26\x20",
    (LPWSTR)"\x1a\x20",
    (LPWSTR)"\x1e\x20",
    (LPWSTR)"\xc6\x02",
    (LPWSTR)"\x39\x20",
    (LPWSTR)"\x18\x20",
    (LPWSTR)"\x19\x20",
    (LPWSTR)"\x1c\x20",
    (LPWSTR)"\x1d\x20",
    (LPWSTR)"\x22\x20",
    (LPWSTR)"\x3a\x20",
    (LPWSTR)"\x62\xff",
    (LPWSTR)"\x63\xff",
    (LPWSTR)"\x64\xff"
  };
static const int ndelimiters = sizeof(delimiters) / sizeof(LPWSTR);

static LPWSTR OthPullJP(void)
{
  int dpos, i;
  LPWSTR newdposptr = NULL;
  LPWSTR retjp = NULL;

  OTHLOCK();

  if (jpendpos == 0)
  {
    goto ret;
  }

  for (i = 0; i < ndelimiters; i++)
  {
    OthStrStr(delimiters[i], &newdposptr);
  }

  if (newdposptr == NULL || newdposptr == jpbuff)
  {
    retjp = NULL;
  }
  else
  {
    dpos = newdposptr - jpbuff + 1;
    retjp = malloc((dpos + 3) << 1);
    retjp[0] = 0x000d;
    retjp[1] = 0x000a;
    memcpy(retjp + 2, jpbuff, dpos << 1);
    retjp[dpos + 2] = 0;

    jpendpos = lstrlenW(jpbuff + dpos);
    memmove(jpbuff, jpbuff + dpos, (jpendpos + 1) << 1);
  }

ret:

  OTHUNLOCK();

  return retjp;
}

static void OthPushKR(LPWSTR krStr)
{
  int i, len;

  if (krStr == NULL)
    return;

  len = lstrlenW(krStr);

  OTHLOCK();

  if (krendpos >= BUFFSIZE)
  {
    krbuff[0] = 0;
    krendpos = 0;
  }

  for (i = 0; i < len && krendpos < BUFFSIZE; ++i, ++krendpos)
  {
    krbuff[krendpos] = krStr[i];
  }

  if (krendpos >= BUFFSIZE)
  {
    krbuff[0] = 0;
    krendpos = 0;
  }

  krbuff[krendpos] = 0;

  OTHUNLOCK();
}

static LPWSTR OthPullKR(void)
{
  LPWSTR ret = NULL;
  int len;

  OTHLOCK();

  if (krendpos != 0)
  {
    len = lstrlenW(krbuff);
    ret = malloc((len + 1) * 2);
    lstrcpyW(ret, krbuff);
    krbuff[0] = 0;
    krendpos = 0;
  }

  OTHUNLOCK();

  return ret;
}

static void OthJKTransThread(void *arg)
{
  LPWSTR jpStr;
  LPWSTR krStr;

  while (isRunning)
  {
    jpStr = OthPullJP();
    if (jpStr != NULL)
    {
      if (jpStr[0] != 0)
      {
        krStr = OthJKTrans(jpStr);
        OthPushKR(krStr);
        free(krStr);
      }
      free(jpStr);
    }
    else
    {
      Sleep(500);
    }
  }
}

OTHEXPORT int OTHCALL OthTransInit(OTHTR *pTR)
{
  int i;
  const OHBSETTING *arrSettings;

  if (isRunning)
  {
    ++isRunning;
    return 0;
  }

  if (pfnOhbInit(OHBID_BABEL_JK) != 0)
  {
    return -1;
  }

  arrSettings = pfnOhbGetSettings(&i);
  if ((i < OHBID_BABEL_JK) ||
    (arrSettings[OHBID_BABEL_JK].iTransID == OHBID_INVALID))
  {
    pfnOhbDestroy(OHBID_BABEL_JK);
    return -2;
  }

  pTransSetting = &arrSettings[OHBID_BABEL_JK];

  pTR->lpStr = NULL;
  pTR->iLen = 0;

  OTHLOCKINIT();

  isRunning = 1;

  jktrth = (HANDLE)_beginthread(&OthJKTransThread, 0, NULL);

  return 0;
}

OTHEXPORT int OTHCALL OthTransSetting(OTHTR *pTR)
{
  return 0;
}

OTHEXPORT int OTHCALL OthTransCopySetting(OTHTR *pDestTR, OTHTR *pSrcTR)
{
  return 0;
}

OTHEXPORT int OTHCALL OthTranslate(OTHTR *pTR, LPCWSTR lpStr, int iLen)
{
  int len;
  LPWSTR krStr;
  LPSTR euckr;

  if (lpStr[0] != 0)
  {
    OthPushJP(lpStr, iLen);
  }

  krStr = OthPullKR();

  free(pTR->lpStr);
  pTR->lpStr = NULL;

  if (krStr)
  {
    pTR->lpStr = krStr;
    pTR->iLen = lstrlenW(krStr);
  }

  if (pTR->lpStr == NULL)
  {
    pTR->lpStr = malloc(2);
    pTR->lpStr[0] = 0;
    pTR->iLen = 0;
  }

  return 0;
}

OTHEXPORT int OTHCALL OthTransDestroy(OTHTR *pTR)
{
  --isRunning;
  if (isRunning != 0)
    return 0;

  if (jktrth)
  {
    WaitForSingleObject(jktrth, INFINITE);
    jktrth = NULL;
  }
  jpbuff[0] = '\0';
  jpendpos = 0;
  krbuff[0] = '\0';
  krendpos = 0;
  OTHLOCKDEL();
  free(pTR->lpStr);
  pfnOhbDestroy(OHBID_BABEL_JK);

  return 0;
}

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{

  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:
      hInst = hinstDLL;
      {
        int i, len;
        TCHAR path[OTHPATHLEN];
        GetModuleFileName(hInst, path, OTHPATHLEN);
        len = lstrlen(path);
        if (len > 2000)
        {
          return FALSE;
        }
        for (i = len - 1; i > 0; --i)
        {
          if (path[i] == (TCHAR)'\\')
          {
            break;
          }
        }
        if (i < 0)
        {
          return FALSE;
        }
        path[i] = 0;
        lstrcat(path, TEXT("\\ohbbhlp\\ohbbhlp.dll"));
        hBabel = LoadLibrary(path);
        if (hBabel == NULL)
        {
          return FALSE;
        }
        pfnOhbInit = (PVOID)GetProcAddress(hBabel, "OhbInit");
        if (pfnOhbInit == NULL)
        {
          return FALSE;
        }
        pfnOhbDestroy = (PVOID)GetProcAddress(hBabel, "OhbDestroy");
        if (pfnOhbDestroy == NULL)
        {
          return FALSE;
        }
        pfnOhbGetSettings = (PVOID)GetProcAddress(hBabel, "OhbGetSettings");
        if (pfnOhbGetSettings == NULL)
        {
          return FALSE;
        }
        pfnOhbTrans = (PVOID)GetProcAddress(hBabel, "OhbTrans");
        if (pfnOhbTrans == NULL)
        {
          return FALSE;
        }
        pfnOhbFree = (PVOID)GetProcAddress(hBabel, "OhbFree");
        if (pfnOhbFree == NULL)
        {
          return FALSE;
        }
      }
      break;

    case DLL_PROCESS_DETACH:
      FreeLibrary(hBabel);
      break;
  }

  return TRUE;
}
